@@ -8,20 +8,23 @@ from shutil import copyfile
 import numpy as np
 from jinja2 import Environment, PackageLoader
 
-from pastml.file import get_pastml_colour_file
 from pastml import numeric2datetime
-from pastml.tree import DATE, DATE_CI
+from pastml.file import get_pastml_colour_file
+from pastml.tree import DATE, IS_POLYTOMY
 from pastml.visualisation import get_formatted_date
 from pastml.visualisation.colour_generator import get_enough_colours, WHITE, parse_colours
 from pastml.visualisation.tree_compressor import NUM_TIPS_INSIDE, TIPS_INSIDE, TIPS_BELOW, \
     REASONABLE_NUMBER_OF_TIPS, compress_tree, INTERNAL_NODES_INSIDE, ROOTS, IS_TIP, ROOT_DATES, IN_FOCUS, AROUND_FOCUS, \
-    UP_FOCUS, IS_POLYTOMY
+    UP_FOCUS
 
 JS_LIST = ["https://pastml.pasteur.fr/static/js/jquery.min.js",
            "https://pastml.pasteur.fr/static/js/jquery.qtip.min.js",
            "https://pastml.pasteur.fr/static/js/cytoscape.min.js",
            "https://pastml.pasteur.fr/static/js/cytoscape-qtip.js",
-           "https://pastml.pasteur.fr/static/js/cytoscape-svg.js"]
+           "https://pastml.pasteur.fr/static/js/cytoscape-svg.js",
+           "https://pastml.pasteur.fr/static/js/layout-base.min.js",
+           "https://pastml.pasteur.fr/static/js/cose-base.min.js",
+           "https://pastml.pasteur.fr/static/js/cytoscape-cose-bilkent.min.js"]
 CSS_LIST = ["https://pastml.pasteur.fr/static/css/jquery.qtip.min.css",
             "https://pastml.pasteur.fr/static/css/bootstrap.min.css"]
 
@@ -41,6 +44,7 @@ UNRESOLVED = 'unresolved'
 TIP = 'tip'
 
 TOOLTIP = 'tooltip'
+COLOUR = 'colour'
 
 DATA = 'data'
 ID = 'id'
@@ -100,9 +104,8 @@ def get_scaling_function(y_m, y_M, x_m, x_M):
 def set_cyto_features_compressed(n, size_scaling, e_size_scaling, font_scaling, transform_size, transform_e_size, state,
                                  root_names, root_dates, suffix='', is_mixed=False):
     tips_inside, tips_below, internal_nodes_inside, roots = \
-        getattr(n, TIPS_INSIDE, []), getattr(n, TIPS_BELOW, []), getattr(n, INTERNAL_NODES_INSIDE, []), getattr(n,
-                                                                                                                ROOTS,
-                                                                                                                [])
+        getattr(n, TIPS_INSIDE, []), getattr(n, TIPS_BELOW, []), \
+        getattr(n, INTERNAL_NODES_INSIDE, []), getattr(n, ROOTS, [])
 
     def get_min_max_str(values, default_value=0):
         min_v, max_v = (min(len(_) for _ in values), max(len(_) for _ in values)) \
@@ -247,8 +250,7 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
                             tips_below_i.append(tb)
                             internal_nodes_inside_i.append([_ for _ in ini if not _.is_leaf()])
                     n.add_features(**{TIPS_INSIDE: tips_inside_i, TIPS_BELOW: tips_below_i,
-                                      INTERNAL_NODES_INSIDE: internal_nodes_inside_i,
-                                      ROOTS: roots_i})
+                                      INTERNAL_NODES_INSIDE: internal_nodes_inside_i, ROOTS: roots_i})
                     if roots_i:
                         n.add_feature(MILESTONE, i)
                         root_names = [getattr(_, BRANCH_NAME) if getattr(_, DATE) > milestone else _.name for _ in
@@ -293,6 +295,17 @@ def _forest2json_compressed(forest, compressed_forest, columns, name_feature, ge
     return json_dict, sorted(clazzes)
 
 
+def binary_search(start, end, value, array):
+    if start >= end - 1:
+        return start
+    i = int((start + end) / 2)
+    if array[i] == value or array[i] > value and (i == start or value > array[i - 1]):
+        return i
+    if array[i] > value:
+        return binary_search(start, i, value, array)
+    return binary_search(i + 1, end, value, array)
+
+
 def _forest2json(forest, columns, name_feature, get_date, milestones=None, timeline_type=TIMELINE_SAMPLED,
                  dates_are_dates=True):
     min_root_date = min(getattr(tree, DATE) for tree in forest)
@@ -325,16 +338,6 @@ def _forest2json(forest, columns, name_feature, get_date, milestones=None, timel
 
     # Save cytoscape features at different timeline points
     if len(milestones) > 1:
-        def binary_search(start, end, value, array):
-            if start >= end - 1:
-                return start
-            i = int((start + end) / 2)
-            if array[i] == value or array[i] > value and (i == start or value > array[i - 1]):
-                return i
-            if array[i] > value:
-                return binary_search(start, i, value, array)
-            return binary_search(i + 1, end, value, array)
-
         for tree in forest:
             for n in tree.traverse('preorder'):
                 ms_i = binary_search(0 if n.is_root() else getattr(n.up, MILESTONE),
@@ -396,6 +399,90 @@ def _forest2json(forest, columns, name_feature, get_date, milestones=None, timel
     return json_dict, sorted(clazzes)
 
 
+def _forest2json_transitions(states, counts, transitions, state2colour, threshold=0):
+    nodes, edges = [], []
+    n = len(states)
+
+    n_scaler = get_scaling_function(y_m=200, y_M=800, x_m=min(counts), x_M=max(counts))
+    font_scaler = get_scaling_function(y_m=MIN_FONT_SIZE, y_M=MIN_FONT_SIZE * 3, x_m=min(counts), x_M=max(counts))
+    positive_transitions = transitions[transitions > 0]
+    e_scaler = get_scaling_function(y_m=30, y_M=200, x_m=min(positive_transitions), x_M=max(positive_transitions))
+
+    transtions_to_from = np.transpose(transitions.copy())
+    np.fill_diagonal(transtions_to_from, 0)
+    nums = np.triu(transitions + transtions_to_from).flatten()
+    positive_nums = nums[nums > 0]
+
+    max_transition_num = max(positive_nums)
+    if max_transition_num <= 2:
+        miles = sorted((set({0, 1 if max_transition_num > 1 else 0}
+                                     | set(np.round(positive_nums, 3))) - {np.round(max_transition_num, 3)}))
+    else:
+        miles = sorted(set({0, 1} | set(np.trunc(positive_nums))))
+    miles = np.array([_ for _ in miles if threshold <= _ < max_transition_num])
+    if not len(miles):
+        miles = [0]
+
+    # we hide connections when they are < mile
+    def get_mile(start, end, value):
+        if start == end:
+            return None
+        i = int((start + end) / 2)
+        if miles[i] <= value:
+            if i == end - 1 or value < miles[i + 1]:
+                return i
+            return get_mile(i + 1, end, value)
+        return get_mile(start, i, value)
+
+    i2mile = defaultdict(lambda: -1)
+    for i in range(n):
+        from_state = states[i]
+        n_tips = counts[i]
+        i_node_size = n_scaler(n_tips)
+        if n_tips > 0:
+            for j in range(i, n):
+                if counts[j] > 0:
+                    to_state = states[j]
+                    n_ij = transitions[i, j]
+                    n_ji = transitions[j, i]
+                    node_size = (i_node_size + n_scaler(counts[j])) / 2
+                    if n_ij > 0:
+                        mile = get_mile(0, len(miles), n_ij)
+                        if mile is not None:
+                            i2mile[i] = max(mile, i2mile[i])
+                            i2mile[j] = max(mile, i2mile[j])
+                            edges.append(get_edge(i, j,
+                                                  **{ID: '{}_{}'.format(i, j),
+                                                     EDGE_SIZE: e_scaler(n_ij),
+                                                     NODE_SIZE: node_size / (2 if i != j else 1),
+                                                     EDGE_NAME: n_ij,
+                                                     TOOLTIP: '{} transitions from {} to {}'
+                                                  .format(n_ij, from_state, to_state),
+                                                     MILESTONE: mile}))
+                    if n_ji > 0 and i != j:
+                        mile = get_mile(0, len(miles), n_ji)
+                        if mile is not None:
+                            i2mile[i] = max(mile, i2mile[i])
+                            i2mile[j] = max(mile, i2mile[j])
+                            edges.append(get_edge(j, i,
+                                                  **{ID: '{}_{}'.format(j, i),
+                                                     EDGE_SIZE: e_scaler(n_ji),
+                                                     NODE_SIZE: node_size / 2,
+                                                     EDGE_NAME: n_ji,
+                                                     TOOLTIP: '{} transitions from {} to {}'
+                                                  .format(n_ji, to_state, from_state),
+                                                     MILESTONE: mile}))
+            if i2mile[i] >= 0:
+                nodes.append(_get_node(data={ID: i, NODE_NAME: from_state,
+                                             NODE_SIZE: i_node_size,
+                                             FONT_SIZE: font_scaler(n_tips),
+                                             TOOLTIP: '{} is represented by {:.0f} samples.'.format(from_state, n_tips),
+                                             COLOUR: state2colour[from_state],
+                                             MILESTONE: i2mile[i]}))
+    json_dict = {NODES: nodes, EDGES: edges}
+    return json_dict, ['{:g}'.format(_) for _ in miles]
+
+
 def get_size_transformations(forest):
     max_size, min_size, max_e_size, min_e_size = 1, np.inf, 1, np.inf
     for tree in forest:
@@ -481,6 +568,7 @@ def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2c
                             else 'diversification events',
                             age_label=milestone_label)
     slider = env.get_template('time_slider.html').render(min_date=0, max_date=len(milestones) - 1,
+                                                         cur_date=len(milestones) - 1,
                                                          name=milestone_label) if len(milestones) > 1 else ''
 
     template = env.get_template('index.html')
@@ -516,6 +604,60 @@ def save_as_cytoscape_html(forest, out_html, column2states, name_feature, name2c
         fp.write(page)
 
 
+def save_as_transition_html(character, states, counts, transitions, out_html, state2colour, work_dir,
+                            local_css_js=False, threshold=0):
+    """
+    Converts transition count data to an html representation using Cytoscape.js.
+
+    :param out_html: path where to save the resulting html file.
+    """
+    graph_name = os.path.splitext(os.path.basename(out_html))[0]
+    transitions[transitions < threshold] = 0
+    json_dict, thresholds = _forest2json_transitions(states, counts, transitions, state2colour, threshold=threshold)
+
+    loader = PackageLoader('pastml')
+    env = Environment(loader=loader)
+    template = env.get_template('transitions.js')
+
+    graph = template.render(elements=json_dict, character=character,
+                            years=thresholds)
+    slider = env.get_template('time_slider.html').render(min_date=0, max_date=len(thresholds) - 1,
+                                                         name='transition number threshold', cur_date=0) if len(
+        thresholds) > 1 else ''
+
+    template = env.get_template('index.html')
+    os.makedirs(os.path.abspath(os.path.dirname(out_html)), exist_ok=True)
+
+    if local_css_js:
+        js_list = []
+        os.makedirs(os.path.join(work_dir, 'js'), exist_ok=True)
+        os.makedirs(os.path.join(work_dir, 'css'), exist_ok=True)
+        os.makedirs(os.path.join(work_dir, 'fonts'), exist_ok=True)
+
+        template_dir = os.path.join(os.path.abspath(os.path.split(__file__)[0]), '..', 'templates')
+        for _ in sorted(glob(os.path.join(template_dir, 'js', '*.js*'))):
+            cp = os.path.join(work_dir, 'js', os.path.split(_)[1])
+            copyfile(_, cp)
+            if cp.endswith('.js'):
+                js_list.append(cp)
+        css_list = []
+        for _ in glob(os.path.join(template_dir, 'css', '*.css*')):
+            cp = os.path.join(work_dir, 'css', os.path.split(_)[1])
+            copyfile(_, cp)
+            if cp.endswith('.css'):
+                css_list.append(cp)
+        for _ in glob(os.path.join(template_dir, 'fonts', '*.*')):
+            cp = os.path.join(work_dir, 'fonts', os.path.split(_)[1])
+            copyfile(_, cp)
+    else:
+        js_list = JS_LIST
+        css_list = CSS_LIST
+    page = template.render(graph=graph, title=graph_name, slider=slider, js_list=js_list, css_list=css_list)
+
+    with open(out_html, 'w+') as fp:
+        fp.write(page)
+
+
 def _clazz_list2css_class(clazz_list):
     if not clazz_list:
         return None
@@ -545,11 +687,25 @@ def get_column_value_str(n, column, format_list=True, list_value=''):
 
 def visualize(forest, column2states, work_dir, name_column=None, html=None, html_compressed=None, html_mixed=None,
               tip_size_threshold=REASONABLE_NUMBER_OF_TIPS, date_label='Dist. to root', timeline_type=TIMELINE_SAMPLED,
-              local_css_js=False, column2colours=None, focus=None, should_resolve_polytomies=False):
+              local_css_js=False, column2colours=None, focus=None):
+    for tree in forest:
+        nodes_in_focus = set()
+        for node in tree.traverse():
+            for column in column2states.keys():
+                col_state = getattr(node, column, set())
+                if focus and col_state & focus[column]:
+                    nodes_in_focus.add(node)
+        for node in nodes_in_focus:
+            node.add_feature(IN_FOCUS, True)
+            if not node.is_root() and node.up not in nodes_in_focus:
+                node.up.add_feature(UP_FOCUS, True)
+            for c in node.children:
+                if c not in nodes_in_focus:
+                    c.add_feature(AROUND_FOCUS, True)
+
     one_column = next(iter(column2states.keys())) if len(column2states) == 1 else None
 
     name2colour = {}
-    state2color = None
     for column, states in column2states.items():
         num_unique_values = len(states)
         if column2colours and column in column2colours:
@@ -578,7 +734,6 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
                                           'and serialized this mapping to {}.'
                                           .format(column, states, colours, out_colour_file))
     for tree in forest:
-        nodes_in_focus = set()
         for node in tree.traverse():
             if node.is_leaf():
                 node.add_feature(IS_TIP, True)
@@ -587,15 +742,7 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
                 col_state = getattr(node, column, set())
                 if len(col_state) != 1:
                     node.add_feature(UNRESOLVED, 1)
-                if focus and col_state & focus[column]:
-                    nodes_in_focus.add(node)
-        for node in nodes_in_focus:
-            node.add_feature(IN_FOCUS, True)
-            if not node.is_root() and node.up not in nodes_in_focus:
-                node.up.add_feature(UP_FOCUS, True)
-            for c in node.children:
-                if c not in nodes_in_focus:
-                    c.add_feature(AROUND_FOCUS, True)
+                    break
 
     if TIMELINE_NODES == timeline_type:
         def get_date(node):
@@ -629,8 +776,6 @@ def visualize(forest, column2states, work_dir, name_column=None, html=None, html
     if milestone_labels is None:
         milestone_labels = ['{:g}'.format(_) for _ in milestones]
 
-    if should_resolve_polytomies:
-        resolve_polytomies(column2states, forest, name_column, state2color)
     if html:
         total_num_tips = sum(len(tree) for tree in forest)
         if total_num_tips > MAX_TIPS_FOR_FULL_TREE_VISUALISATION:
@@ -696,54 +841,3 @@ def update_milestones(forest, date_label, milestone_labels, milestones, timeline
     if milestone_labels is None:
         milestone_labels = ['{:g}'.format(_) for _ in milestones]
     return milestone_labels, milestones
-
-
-def resolve_polytomies(column2states, forest, name_column, state2color):
-    columns = sorted(column2states.keys())
-
-    col2state2i = {c: dict(zip(states, range(len(states)))) for (c, states) in column2states.items()}
-
-    def get_prediction(n):
-        return '.'.join('-'.join(str(i) for i in sorted([col2state2i[c][_] for _ in getattr(n, c, set())])) for c in columns)
-
-    for tree in forest:
-        todo = [tree]
-        while todo:
-            n = todo.pop()
-            n_state = get_prediction(n)
-            todo.extend(n.children)
-            if len(n.children) > 2:
-                state2children = defaultdict(list)
-                for c in n.children:
-                    state2children[get_prediction(c)].append(c)
-                for state, children in state2children.items():
-                    state_change = state != n_state
-                    if (state_change or getattr(n, UP_FOCUS, False)) and len(children) > 1:
-                        child = min(children, key=lambda _: _.dist)
-                        dist = child.dist if state_change else 0
-                        pol = n.add_child(dist=dist, name='{}.polytomy_{}'.format(n.name, state))
-                        pol.add_feature(IS_POLYTOMY, 1)
-                        pol.add_feature(DATE, getattr(child, DATE) if state_change else getattr(n, DATE))
-                        pol.add_feature(DATE_CI, getattr(child, DATE_CI, None) if state_change else getattr(n, DATE_CI, None))
-                        for c in columns:
-                            pol.add_feature(c, getattr(child, c))
-                        if hasattr(child, UNRESOLVED):
-                            pol.add_feature(UNRESOLVED, getattr(child, UNRESOLVED))
-                        if hasattr(child, IN_FOCUS):
-                            pol.add_feature(IN_FOCUS, getattr(child, IN_FOCUS))
-                        if hasattr(child, AROUND_FOCUS):
-                            pol.add_feature(AROUND_FOCUS, getattr(child, AROUND_FOCUS))
-                        if name_column is not None:
-                            sts = getattr(pol, name_column, set())
-                            if len(sts) == 1:
-                                col = state2color[next(iter(sts))]
-                                if getattr(n, name_column, set()) == sts:
-                                    pol.add_feature('edge_color', col)
-                                for c in children:
-                                    c.add_feature('edge_color', col)
-                        pol.add_feature(BRANCH_NAME, '{}-{}'.format(n.name, pol.name))
-                        for c in children:
-                            c.add_feature(BRANCH_NAME, '{}-{}'.format(pol.name, c.name))
-                        for c in children:
-                            n.remove_child(c)
-                            pol.add_child(c, dist=c.dist - dist)
